// Downscale a 24-bit TGA image using the Lanczos3 filter in linear color space.
// Usage: downscale image.tga factor

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>

#define die(...) ( fprintf(stderr, __VA_ARGS__), exit(-1), 1 )
#define alloc(p, elems) ( ((p) = calloc((elems), sizeof(*(p)))) || die("Allocation error!") )


int width, height, decimation;

#define OUTPUT_WIDTH ((width+decimation-1)/decimation)
#define OUTPUT_HEIGHT ((height+decimation-1)/decimation)
#define CHANNELS 3

#define FILTER_LENGTH (decimation*6 + (decimation&1))
#define FILTER_OFFSET (decimation/2 - 3*decimation)

#define CACHED_ROWS FILTER_LENGTH

// Helpers.

#define PI 3.14159265358979323846

#ifndef max
  #define max(a,b) ((a)>(b) ? (a) : (b))
#endif
#ifndef min
  #define min(a,b) ((a)<(b) ? (a) : (b))
#endif


// Gamma correction.

double sRGB255_to_linearRGB1_table[256];  // speedup
void compute_sRGB255_to_linearRGB1_table(void) {
  int i;
  for (i=0; i<256; i++) {
    double x = i/255.0;
    sRGB255_to_linearRGB1_table[i] =
      x <= 0.04045 ? x / 12.92 : pow((x + 0.055) / 1.055, 2.4);
  }
}
double sRGB255_to_linearRGB1(unsigned char x) {
  return sRGB255_to_linearRGB1_table[x];
}

unsigned char linearRGB1_to_sRGB255(double l) {
  if (l <= 0) return 0;
  else if (l >= 1) return 255;
  else if (l <= 0.0031308) return floor(0.5 + 255 * (l * 12.92));
  else return floor(0.5 + 255 * (pow(l, 1.0/2.4) * 1.055 - 0.055));
}


// The Lanczos3 filter.
double* filterCoef;
double lanczos3(double x) {
  if (x == 0) return 1;
  else return 3 * sin(PI*x) * sin(PI*x/3) / (PI*PI*x*x);
}

void compute_filterCoef(void) {
  alloc(filterCoef, FILTER_LENGTH);

  int i;
  if (decimation & 1) {  // odd filter
    for (i=0; i<FILTER_LENGTH; i++) {
      filterCoef[i] = lanczos3(-3 + (double)i/decimation);
    }
  }
  else {  // even filter
    for (i=0; i<FILTER_LENGTH; i++) {
      filterCoef[i] = lanczos3(-3 + (i+0.5)/decimation);
    }
  }
}


double* yScale;
double* xScale;

// Compute filter scale factors (all taps should add to 1).
void compute_scales(void) {
  alloc(yScale, OUTPUT_HEIGHT);
  alloc(xScale, OUTPUT_WIDTH);

  int outCol, col, outRow, row, ch, i;
  for (outCol=0; outCol<OUTPUT_WIDTH; outCol++) {
    int startCol = outCol*decimation + FILTER_OFFSET;
    int firstCol = max(0, startCol);
    int lastCol = min(startCol + FILTER_LENGTH - 1, width-1);
    double scale = 0;
    for (col=firstCol, i=col-startCol; col<=lastCol; col++, i++) {
      scale += filterCoef[i];
    }
    xScale[outCol] = 1./scale;
  }
  for (outRow=0; outRow<OUTPUT_HEIGHT; outRow++) {
    int startRow = outRow*decimation + FILTER_OFFSET;
    int firstRow = max(0, startRow);
    int lastRow = min(startRow + FILTER_LENGTH - 1, height-1);
    double scale = 0;
    for (row=firstRow, i=row-startRow; row<=lastRow; row++, i++) {
      scale += filterCoef[i];
    }
    yScale[outRow] = 1./scale;
  }
}


int lastCachedRow = INT_MIN;

// Read an image row and transform it into linearized RGB.
void readRow(double* dst, FILE* fr, int rowNum) {
  int x, ch;
  unsigned char row[width][CHANNELS];

  if (lastCachedRow < rowNum) {
    fseek(fr, 18 + rowNum*width*CHANNELS, SEEK_SET);
    fread(&row[0][0], 1, width*CHANNELS, fr);

    for (x=0; x<width; x++) for (ch=0; ch<CHANNELS; ch++) {
      dst[x*CHANNELS+ch] = sRGB255_to_linearRGB1(row[x][ch]);
    }
  }
  lastCachedRow = max(lastCachedRow, rowNum);
}


double (*input)[CHANNELS], (*afterY)[CHANNELS];


// Main function.
int main(int argc, char** argv) {
  if (argc != 3) die("Usage: downscale image.tga factor");
  decimation = atoi(argv[2]); if (decimation <= 0) die("Usage: downscale image.tga factor");

  // Open files.
  FILE *fr = fopen(argv[1], "rb"); if (!fr) die("Couldn't open \"%s\"!", argv[1]);
  unsigned char rHeader[18];
  fread(rHeader, 18, 1, fr);

  if (rHeader[0] || rHeader[1] || rHeader[2]!=2 || rHeader[16]!=24) {
    die("The only supported TGA format is uncompressed 24-bit.");
  }

  width = rHeader[12] + rHeader[13]*256;
  height = rHeader[14] + rHeader[15]*256;


  char newFile[1024]; sprintf(newFile, "dec%02d_%s", decimation, argv[1]);
  FILE *fw = fopen(newFile, "wb"); if (!fw) die("Couldn't open output file.");

  // Init.
  compute_sRGB255_to_linearRGB1_table();
  compute_filterCoef();
  compute_scales();

  alloc(input, CACHED_ROWS * width);
  alloc(afterY, width + FILTER_LENGTH); afterY = &afterY[-FILTER_OFFSET];

  // Write header. Keep orientation flags.
  unsigned char header[18] = {
    0,0,2,0,0, 0,0,0,0,0, 0,0,
    OUTPUT_WIDTH%256,OUTPUT_WIDTH/256,
    OUTPUT_HEIGHT%256,OUTPUT_HEIGHT/256,
    24,rHeader[17]&0x30
  };
  fwrite(header, 18, 1, fw);

  // Write filtered image.
  int outRow, row, outCol, col, i, x, ch;

  for (outRow=0; printf("\r%d", outRow), outRow<OUTPUT_HEIGHT; outRow++) {
    int startRow = outRow*decimation + FILTER_OFFSET;
    int firstRow = max(0, startRow);
    int lastRow = min(startRow + FILTER_LENGTH - 1, height-1);

    // Filter along Y.
    for (x=0; x<width; x++) for (ch=0; ch<CHANNELS; ch++) {
      afterY[x][ch] = 0;
    }
    for (row=firstRow, i=row-startRow; row<=lastRow; row++, i++) {
      int r = (row+CACHED_ROWS) % CACHED_ROWS;
      readRow(&input[r*width][0], fr, row);

      for (x=0; x<width; x++) for (ch=0; ch<CHANNELS; ch++) {
        afterY[x][ch] += input[r*width+x][ch] * filterCoef[i];
      }
    }

    // Filter along X and output the result row.
    for (outCol=0; outCol<OUTPUT_WIDTH; outCol++) {
      int startCol = outCol*decimation + FILTER_OFFSET;
      int firstCol = startCol;
      int lastCol = startCol + FILTER_LENGTH - 1;

      for (ch=0; ch<CHANNELS; ch++) {
        double sum = 0;
        for (col=firstCol, i=col-startCol; col<=lastCol; col++, i++) {
          sum += afterY[col][ch] * filterCoef[i];
        }
        putc(linearRGB1_to_sRGB255(sum * xScale[outCol] * yScale[outRow]), fw);
      }
    }
  }

  fclose(fr);
  fclose(fw);

  return 0;
}
